#######################################################################################################################
# Эмперические характеристики
#######################################################################################################################
def EC_1(marks_text):
    """В группе учится `N` студентов.
    Ими были получены следующие 100-балльные оценки: `marks_text`
    Найдите: 
    - A – среднюю положительную оценку в группе;
    - M – медиану положительных оценок в группе;
    - H – среднее гармоническое оценок, которые не менее M;
    - G – среднее геометрическое оценок, которые не менее M;
    - Q – медианную оценку в той части группы, в которой студенты набрали не менее M баллов;
    - N – количество студентов, оценки которых оказались между H и Q (включая границы)

    Args:
        marks_text (numerical): Строка оценок с запятыми и ковычками

    ## Prints
        `answer` каждое значение по очереди.<br>C запятой вместо точки и сокращением до соответствующего количества знаков после запятой

    Returns:
        `answer` (tuple): Соответствующие величины
    """

    import numpy as np
    import scipy.stats
    
    marks_0 = np.array(marks_text.split(', ')).astype(int)
    marks = marks_0[marks_0!=0]
    
    answer = []
    
    answer.append(np.mean(marks))
    print('A = ' + rrstr(answer[0],1))
    
    answer.append(np.median(marks))
    print('M = ' + str(int(answer[1])))
    
    answer.append(scipy.stats.hmean(marks[marks >= answer[1]]))
    print('H = ' + rrstr(answer[2],1))
    
    answer.append(scipy.stats.gmean(marks[marks >= answer[1]]))
    print('G = ' + rrstr(answer[3],1))
    
    answer.append(np.median(marks[marks >= answer[1]]))
    print('Q = ' + rrstr(answer[4],0))
    
    answer.append(len(marks[[i in np.arange(min(int(answer[2]),int(answer[4])),max(int(answer[2]),int(answer[4]))+1,1) for i in marks]]))
    print('N = ' + rrstr(answer[5],0))
    
    return tuple(answer)
#######################################################################################################################
def EC_2(text):
    """Следующие `N` чисел – это умноженные на 10000 и округленные до ближайшего целого
    дневные логарифмические доходности акции компании АВС: `text`. 
    Будем называть их преобразованными доходностями (ПД).
    Финансовый аналитик Глеб предполагает, что преобразованные доходности (как и исходные)
    приближенно распределены по нормальному закону.
    Чтобы проверить свое предположение Глеб нашел нижнюю квартиль L и верхнюю квартиль H нормального распределения N(μ,σ2),
    для которого μ – это среднее арифметическое ПД,
    а σ – эмпирическое стандартное отклонение ПД.
    Затем Глеб подсчитал количество ПД, попавших в интервал от L до H 
    (надеясь, что в этот интервал попадет половина ПД).
    Результат этого вычисления показался ему недостаточно убедительным.
    Чтобы окончательно развеять сомнения относительно нормальности ПД,
    Глеб построил на одном рисунке графики функций:
    F^(x) и F(x), где F^(x) – эмпирическая функция распределения ПД,
    а F(x) – функция распределения N(μ,σ2).
    В качестве меры совпадения двух графиков Глеб решил использовать
    расстояние d между функциями F^(x) и F(x),
    которое он вычислил, исходя из определения:
    d=sup|F^(x)−F(x)|. В ответе укажите результаты вычислений Глеба: 
    - среднее арифметическое ПД; 
    - эмпирическое стандартное отклонение ПД; 
    - квартили L;
    - квартиль H 
    - количество ПД, попавших в интервал от L до H; 
    - расстояние между функциями F^(x) и F(x).

    Args:
        text (str): Строка чисел через запятую

    ## Prints
        `answer` каждое значение по очереди.<br>C запятой вместо точки и сокращением до соответствующего количества знаков после запятой

    Returns:
        `answer` (tuple): Соответствующие величины
    """
    import numpy as np
    import scipy.stats
    import matplotlib.pyplot as plt
    from statsmodels.distributions.empirical_distribution import ECDF
    
    data= np.array(text.split(', ')).astype(int)
    mu=data.mean()
    sigma = data.std()

    X=scipy.stats.norm(mu,sigma)

    L = X.ppf(0.25)
    H = X.ppf(0.75)

    lenLH = len(data[np.logical_and(data>L,data<H)])


    Fexp = ECDF(data)

    fig, ax = plt.subplots(figsize=(6,5))
    xx=np.linspace(X.ppf(0.01),X.ppf(0.99),1000)
    ax.plot(xx,X.cdf(xx))
    ax.plot(xx,Fexp(xx))
    plt.show()
    data_sort=np.sort(data)

    length_ff=max([max(np.abs((i+1)/len(data)-X.cdf(data_sort[i])),np.abs((i)/len(data)-X.cdf(data_sort[i]))) for i in range(len(data))]) # equals to scipy.stats.kstest(data,X.cdf)[0]

    task = 'среднее арифметическое ПД ,эмпирическое стандартное отклонение ПД ,квартиль L,квартиль Н,количество ПД от L до H,Расстояние мужду функциями распределения'.split(',')
    answer=[mu,sigma,L,H,lenLH,length_ff]
    commas = [1,1,1,1,0,3]
    
    for i in range(6):
        print(task[i]+' = '+rrstr(answer[i],commas[i]))
        
    return tuple(answer)
#######################################################################################################################
def EC_3(text):
    """В группе Ω учатся студенты: ω1,...,ω`n`.
    Пусть X и Y – 100-балльные экзаменационные оценки
    по математическому анализу и теории вероятностей.
    Оценки студента ωi обозначаются: xi=X(ωi) и yi=Y(ωi),
    i=1,...,`n`. Все оценки известны: `text`.
    Требуется найти следующие условные эмпирические характеристики:
    - ковариацию X и Y при условии, что одновременно X⩾50 и Y⩾50;
    - коэффициент корреляции X и Y при том же условии.

    Args:
        text (str): Текст вида `xi=X(ωi), yi=Y(ωi),...` записанный в `'''_'''`

    ## Prints
        `answer` каждое значение по очереди.<br>C запятой вместо точки и сокращением до соответствующего количества знаков после запятой

    Returns:
        `answer` (tuple): Соответствующие величины
    """

    import re
    import numpy as np

    x_values = list(map(int, re.findall(r'x\d+=(\d+)', text)))
    y_values = list(map(int, re.findall(r'y\d+=(\d+)', text)))

    # Отфильтровать пары (X, Y), где X >= 50 и Y >= 50
    filtered_data = [(x, y) for x, y in zip(x_values, y_values) if x >= 50 and y >= 50]
    x_filtered, y_filtered = zip(*filtered_data)  # Разделяем на списки x и y

    # Рассчитать ковариацию
    cov_xy = np.cov(x_filtered, y_filtered, ddof=0)[0, 1]

    # Рассчитать коэффициент корреляции
    corr_xy = np.corrcoef(x_filtered, y_filtered)[0, 1]

    # Вывод результатов
    print('Ковариация = ' + rrstr(cov_xy,1))
    print('Коэффициент корреляции = ' + rrstr(corr_xy,3))
    
    return (cov_xy,corr_xy)
#######################################################################################################################
def EC_4(text):
    """Поток Ω состоит из k групп: Ω1,...,Ωk, k=3.
    На потоке учатся n=n1+...+nk студентов,
    где ni – число студентов в группе Ωi, i=1,...,k.
    Пусть X(ω)– 100-балльная оценка студента ω∈Ω.
    Далее используются следующие обозначения:
    x¯i – среднее значение,
    σi – (эмпирическое) стандартное отклонение признака X на группе Ωi.
    Дано: ni=__,x¯i=__, σi=__,
    Требуется найти: 
    - среднее значение X на потоке Ω;
    - (эмпирическое) стандартное отклонение X на потоке Ω.

    Args:
        text (str): Текст всего задания записанный в `'''_'''`
        
    ## Prints
        `answer` каждое значение по очереди.<br>C запятой вместо точки и сокращением до соответствующего количества знаков после запятой

    Returns:
        `answer` (tuple): Соответствующие величины
    """
    import re
    import numpy as np

    x_values = np.array(list(map(int, re.findall(r'x[¯]{0,3}\d+=(\d+)', text))))
    s_values = np.array(list(map(int, re.findall(r'σ\d+=(\d+)', text))))
    n_values = np.array(list(map(int, re.findall(r'n\d+=(\d+)', text))))

    N = sum(n_values)
    X_ = sum(np.multiply(n_values,x_values))/N

    S = np.sqrt(np.sum(n_values*(s_values**2 + (x_values-X_)**2))/N)

    print('Среднее значение = ' + rrstr(X_,1))
    print('Стандартное отклонение = ' + rrstr(S,3))

    return (X_,S)
#######################################################################################################################
EC=[EC_1, EC_2, EC_3, EC_4]


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
#######################################################################################################################
# Модификации модели классификации
#######################################################################################################################
class SVC1:
    """Классификация методом опорных векторов"""
    def __init__(self) -> None:
        """Классификация методом опорных векторов"""
        pass
    
    def predict(self, X):
        """Предсказывает значения эндогенной переменной(Y)

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
            
        Returns:
            pandas.DataFrame: Предсказанные значения эндогенной переменной(Y)
        """
        return np.sign(X @ self.w)
    
    def error(self, X, y):
        """Считает значение ошибки

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
            y (pandas.DataFrame): Эндогенная переменная (Y)

        Returns:
            numerical: error
        """
        return (1 - self.predict(X) * y).sum()
    
    def fit(self, X, y, a=0.1, n=1000):
        """Функция обучения модели.

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
            y (pandas.DataFrame): Эндогенная переменная (Y)
            a (float, optional): Скорость обучения. Defaults to 0.1.
            n (int, optional): Количество итераций(Шагов). Defaults to 1000.
            
        Returns:
            
        """
        self.w = pd.DataFrame(np.ones((X.shape[1], 1)))                 # Вектор-столбец весов
        
        for i in range(n):
            pred = self.predict(X)                                      # Предсказывает класс объектов в итерации
            M_ = 1 - pred * y                                           # Отступ от разделяющей линии
            wrong_preds = (M_[M_ > 0].dropna()).index                   # Индексы неправильных предсказаний
            grad = -1 * (y.iloc[wrong_preds].T @ X.iloc[wrong_preds])   # Вычисляет градиент функции ошибки для итерации
            self.w -= (a * grad).T                                      # Обновление весов
            
            
    def accuracy_score(self, y, y_):
        """Вычисляет accuracy-score

        Args:
            y (pandas.DataFrame): Истинные значения эндогенной переменной (Y-True)
            y_ (pandas.DataFrame): Предсказанные значения эндогенной переменной (Y-Pred)

        Returns:
            numerical: Accuracy-score
        """
        return ((y_ == y).sum()/y.shape[0])[0]
    
    
    def plot(self, X, y):
        """Строит график

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
            y (pandas.DataFrame): Эндогенная переменная (Y)
        """
        h = 0.02
        x_min, x_max = X[0].min() - 1, X[0].max() + 1
        y_min, y_max = X[1].min() - 1, X[1].max() + 1
        xx, yy = np.meshgrid(np.arange(x_min, x_max, h),
                            np.arange(y_min, y_max, h))

        Z_linear = self.predict(np.c_[xx.ravel(), yy.ravel()])
        Z_linear = Z_linear.values.reshape(xx.shape)
        plt.contourf(xx, yy, Z_linear, alpha=0.1)
        plt.scatter(X[0], X[1], c=y)
        plt.title('SVC')
        plt.xlabel('Sepal Length')
        plt.ylabel('Sepal Width')
        plt.xlim(xx.min(), xx.max())
        plt.ylim(yy.min(), yy.max())
        plt.show()
#######################################################################################################################
class LogReg1:
    """Логистическая регрессия Вариант 1
    """
    def __init__(self) -> None:
        """Логистическая регрессия"""
        pass
    
    def predict_prob(self, X):
        """Предсказывает вероятности значения эндогенной переменной(Y)

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
            
        Returns:
            pandas.DataFrame: Предсказанные вероятности значения эндогенной переменной(Y)
        """
        with np.errstate(over='ignore'):
            result = 1/(1 + np.exp(-(X @ self.w)))
        return result
    
    def predict(self, X, wall):
        """Предсказывает значения эндогенной переменной(Y)

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
            
        Returns:
            pandas.DataFrame: Предсказанные значения эндогенной переменной(Y)
        """
        prob = self.predict_prob(X)
        prob[prob > wall] = 1
        prob[prob < wall] = -1
        return prob
    
    def error(self, X, y):
        """Функция ошибки

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
            y (pandas.DataFrame): Эндогенная переменная (Y)

        Returns:
            numerical: значение функции ошибки
        """
        return -(y * np.log(self.predict_prob(X)) + (1 - y) * np.log(self.predict_prob(-X))).sum()
    
    def fit(self, X, y, a=0.1, n=1000):
        """Функция обучения модели.

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
            y (pandas.DataFrame): Эндогенная переменная (Y)
            a (float, optional): Скорость обучения. Defaults to 0.1.
            n (int, optional): Количество итераций(Шагов). Defaults to 1000.
        """
        self.w = pd.DataFrame(np.ones((X.shape[1], 1)))     # Вектор-столбец весов
        for i in range(n):
            grad = -(X.T @ (y - self.predict_prob(X)))      # Вычисляет градиент функции ошибки для итерации
            self.w -= a * grad                              # Обновление весов
            
    def accuracy_score(self, y, y_):
        """Вычисляет accuracy-score

        Args:
            y (pandas.DataFrame): Истинные значения эндогенной переменной (Y-True)
            y_ (pandas.DataFrame): Предсказанные значения эндогенной переменной (Y-Pred)

        Returns:
            numerical: Accuracy-score
        """
        return ((y == y_).sum()/y.shape[0])[0]
    
    def plot(self, X, y):
        """Строит график

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
            y (pandas.DataFrame): Эндогенная переменная (Y)
        """
        h = 0.02
        x_min, x_max = X[0].min() - 1, X[0].max() + 1
        y_min, y_max = X[1].min() - 1, X[1].max() + 1
        xx, yy = np.meshgrid(np.arange(x_min, x_max, h),
                            np.arange(y_min, y_max, h))

        Z_linear = self.predict(np.c_[xx.ravel(), yy.ravel()], wall=0.5)
        Z_linear = Z_linear.values.reshape(xx.shape)
        plt.contourf(xx, yy, Z_linear, alpha=0.1)
        plt.scatter(X[0], X[1], c=y)
        plt.title('Logistic Regression')
        plt.xlabel('Sepal Length')
        plt.ylabel('Sepal Width')
        plt.xlim(xx.min(), xx.max())
        plt.ylim(yy.min(), yy.max())
        plt.show()
#######################################################################################################################
def sigm(x):
    """Векторизованная функция сигмоиды"""
    with np.errstate(over='ignore'):
        return 1 / (1 + np.exp(-1*x))

class LogReg2():
    """Логистическая регрессия Вариант 2"""
    def __init__(self,X,Y):
        """Логистическая регрессия Вариант 2

        Args:
            X (array-like): Экзогенная переменная (X)
            Y (array-like): Эндогенная переменная (Y)
        """
        self.X = np.matrix(X)
        self.Y = np.matrix(Y).T
        self.b = np.matrix(np.ones(X.shape[1])).T

    def grad(self):
        """Вычисление градиента в данный момент"""
        s = sigm(self.X*self.b)
        return self.X.T * (self.Y - s)

    def fit(self,alpha=0.5, iters=1000):
        """Функция обучения

        Args:
            alpha (float, optional): Скорость обучения. Defaults to 0.5.
            iters (int, optional): Количество итераций(Шагов). Defaults to 1000.
        """
        for i in range(iters):
            self.b+=alpha*self.grad()
            alpha/=2

    def predict(self,X):
        """Предсказание классов объектов (X)"""
        return np.where(np.sign(np.matrix(X)*self.b)>0,1,0).T

    def predict_proba(self,X):
        """Предсказание вероятности принадлежности к классам объектов (X)"""
        return (np.matrix(X)*self.b).T
#######################################################################################################################
class SVC2():
    """Классификация методом опорных векторов с определением ядра"""
    def K(self,X, func = lambda X : X * X.T):
        """Преобразование экзогенной переменной с помощью ядерной функции

        Args:
            X (array-like): Экзогенная переменная (X)
            func (func, optional): Функция преобразования. Defaults to lambda X : X*X.T
            
        Returns:
            (array-like): Экзогенная переменная (X)
        """
        return func(X)

    def __init__(self,X,Y,lamb = 0):
        """Классификация методом опорных векторов с определением ядра

        Args:
            X (array-like): Экзогенная переменная (X)
            Y (array-like): Эндогенная переменная (Y)
            lamb (int, optional): Параметр регуляризации. Defaults to 0.
        """
        self.X = self.K(np.matrix(X))
        self.Y = np.matrix(Y).T
        self.b = np.matrix(np.ones(self.X.shape[1])).T
        self.lamb = lamb

    def predict(self,X):
        """Предсказывает значения эндогенной переменной(Y)

        Args:
            X (array-like): Экзогенная переменная (X)

        Returns:
            array-like: Предсказанные значения эндогенной переменной(Y)
        """
        xs = np.sign(self.K(np.matrix(X)) * self.b)
        return np.where(xs>0,1,0)
    
    def M_crit(self):
        """Вычисление Margin - значения

        Returns:
            array-like: Margin
        """
        nz = np.zeros(1)
        crit =  1 - (self.X*self.b).T * self.Y
        real = (-1)* self.X.T*self.Y
        return np.where(crit>0,real,nz)
    
    def L2_grad(self):
        """Градиент члена функции потерь с регуляризацией"""
        return 2*self.lamb*self.b
    
    def grad(self):
        """Общий градиент функции"""
        self.M_crit()
        return self.M_crit() + self.L2_grad()

    def fit(self,alpha=0.5,iters=1000):
        """Функция обучения"""
        for i in range(iters):
            self.b-=alpha*self.grad()
            alpha/=2
#######################################################################################################################
class KNN:
    def p(self, u, x):
        """Вычисление Евклидова расстояния"""
        return np.sqrt(((u - x)**2).sum())
    def kernel(self, x):
        """Вычисление ядерной функции

        Args:
            x (numerical, array-like): аргумент функции ядра

        Returns:
            (numerical, array-like): Значение ядерной функции
        """
        I = ['Rectangular','Triangular','Epanechnikov','Quartic','Triweight','Tricube','Cosine']
        if (self.kernel_type in I):
            assert np.abs(x)<=1, 'Ядро не подходит для данных' 
            
        kernel = {
            'Rectangular': 1/2,         # Оно же Юниформ или Нормальнок распределение
            'Triangular': (1-np.abs(x)),
            'Epanechnikov' :  3/4 * (1- x**2),
            'Quartic' : 15/16 * (1-x**2)**2,            #Оно же биквадратное
            'Triweight' : 35/32 * (1-x**2)**3,
            'Tricube':  70/81 * (1 - np.abs(x)**3)**3,
            'Gaussian': 1/np.sqrt(2*np.pi) * np.e**(-2*x**2),
            'Cosine': np.pi/4 * np.cos(np.pi/2 * x),
            'Logistic' : 1/(np.e**x + 2 + np.e**(-x)),
            'Sigmoid' : 2/np.pi * 1/(np.e**x + np.e**(-x)),
            'Silverman' : 1/2 * np.e**(-np.abs(x)/np.sqrt(2))*np.sin(np.abs(x)/np.sqrt(2) + np.pi/4),
            None: x     
        }
        return kernel[self.kernel_type]
    
    def predict(self, X, k=3, h=0.1):
        """Предсказывает значения эндогенной переменной(Y)

        Args:
            X (array-like): Экзогенная переменная (X)
            k (int,optional): Количество рассматриваемых соседей
            h (float, optional): Гиперпараметр для ядерной функции

        Returns:
            array-like: Предсказанные значения эндогенной переменной(Y)
        """
        p = {}  # Матрица расстояний
        index = 0
        for obj in X.values:    # Для каждого объекта из тех, которые мы хотим определить, создаем список расстояний до всех точек обучающей выборки
            p_obj = []
            for edu_obj in self.X.values:
                p_obj.append(self.p(obj, edu_obj))
            p[f'{index}'] = p_obj
            index += 1
        p = pd.DataFrame(p)
        ans = []
        for name_col in p.columns:      # По матрица находим класс точек наиболее близких к целевым
            idx = np.argpartition(p[name_col], k)[:k]           # Индексы k точек с минимальным расстояним
            class_y = [self.y.iloc[obj].values[0] for obj in idx]       # Каждой ближайшей точке находим соответствующий класс
            if self.kernel_type:
                kernel_k_near = [self.kernel(obj/h) for obj in p[name_col][idx].values] # Значение ядерной функции для каждой точки
                y_w = pd.DataFrame({'Y':class_y, 'w':kernel_k_near}) # Группирую по классу, нахожу сумму весов для каждого и выбираю больший
                ans.append(y_w.groupby('Y').sum('w').idxmax().values[0])
            else:
                vals, count = np.unique(class_y, return_counts=True)
                ans.append(vals[np.argmax(count)]) # Находим наиболее встречающийся класс                
                
        return pd.DataFrame(ans)
    
    def fit(self, X, y, kernel_type = 'Gaussian'):
        """Добавим тренировочную выборку в модель и определим тип ядра из
        `[
            'Rectangular',
            'Triangular',
            'Epanechnikov',
            'Quartic',
            'Triweight',
            'Tricube',
            'Gaussian',
            'Cosine',
            'Logistic',
            'Sigmoid',
            'Silverman',
            None
        ]`. При значении `None` будет применен невзвешенный k-NN
        
        Args:
            X (array-like): Экзогенная переменная (X)
            Y (array-like): Эндогенная переменная (Y)
        """
        self.X = X
        self.y = y
        self.kernel_type = kernel_type
        self.data = pd.concat([X, y], axis=1)
        return self
    
    def accuracy_score(self, y, y_):
        """Вычисление accuracy-значения модели

        Args:
            y (array-like): Истинные значения
            y_ (array-like): Предсказанные значения

        Returns:
            float: accuracy-score
        """
        return ((y_ == y).sum()/y.shape[0])[0]
    
    def plot(self, X, y):
        """Диаграмма рассеяния точек, где 'o' - обучающая выборка, а 'x' - целевая выборка

        Args:
            X (array-like): Экзогенная переменная целевой выборки (X)
            y (array-like): Эндогенная переменная целевой выборки (Y)
        """
        plt.scatter(self.X[0], self.X[1], c=self.y, marker='o') # Скаттер обучающей выборки
        plt.scatter(X[0], X[1], c=y, marker='x') # Скаттер целевых точек
        plt.show()
#######################################################################################################################        
CM = [LogReg1,LogReg2,SVC1,SVC2,KNN]


#######################################################################################################################
# Приближенное вычисление вероятности методом Монте-Карло
#######################################################################################################################
def ACMK_1(a,b,le1,le2):
    """В прямоугольной области, заданной ограничениями |x| ⩽ `a` и |y| ⩽ `b`,
    случайным образом выбираются две точки: (x1,y1) и (x2,y2).
    Пусть A и B - события, состоящие в том, что: 
    A - расстояние между выбранными точками меньше `le1`;
    B - модуль разности |x1 - x2| меньше `le2`.
    Найдите приближенно, методом Монте-Карло:
    - вероятность P(A)
    - условную вероятность P(A|B)
    Указание: получите в заданной прямоугольной области 100000 пар точек и, используя все эти точки, найдите ответы, округляя их до одного знака после запятой.

    Args:
        a (numerical): Значение меньше которого модуль X
        b (numerical): Значение меньше которого модуль Y
        le1 (numerical): Число из первого(A) условия
        le2 (numerical): Число из второго(B) условия

    ## Prints
        `answer` каждое значение по очереди.<br>C запятой вместо точки и сокращением до соответствующего количества знаков после запятой

    Returns:
        `answer` (tuple): Соответствующие величины
    """
    import scipy.stats
    
    # Вероятность P(A)
    X = scipy.stats.uniform(0, 2*a)
    Y = scipy.stats.uniform(0, 2*b)
    N = 100_000
    count = 0
    for i in range(N):
        x1 = X.rvs(size=1)[0]
        y1 = Y.rvs(size=1)[0]
        x2 = X.rvs(size=1)[0]
        y2 = Y.rvs(size=1)[0]
        if ((x2 - x1)**2 + (y2 - y1)**2)**0.5 < le1:
            count += 1
            
    answer = [count/N]
    print('Вероятность P(A) = ' + rrstr(answer[0],1))
    
    # Вероятность P(A|B)
    X = scipy.stats.uniform(0, 2*a)
    Y = scipy.stats.uniform(0, 2*b)
    N = 100_000
    count1 = 0
    count2 = 0
    for i in range(N):
        x1 = X.rvs(size=1)[0]
        y1 = Y.rvs(size=1)[0]
        x2 = X.rvs(size=1)[0]
        y2 = Y.rvs(size=1)[0]
        if ((x2 - x1)**2 + (y2 - y1)**2)**0.5 < le1 and abs(x2 - x1) < le2:
            count1 += 1
        if abs(x2 - x1) < 14:
            count2 += 1
            
    answer.append(count1/count2)
    print('Условную вероятность P(A|B) = ' + rrstr(answer[1],1))
    
    answer = tuple(answer)
    return answer
#######################################################################################################################
def ACMK_2(a,b,le,sign):
    """В области, ограниченной эллипсом (u/`a`)^2+(v/`b`)^2=1,
    случайным образом выбираются две точки.
    Пусть A и B – события, состоящие в том, что:
    A - расстояние между выбранными точками меньше `le`;
    B - все координаты обеих точек `sign` 0. 
    Найдите приближенно, методом Монте-Карло: 
    - вероятность P(A)
    - условную вероятность P(A|B)
    
    Указание: получите внутри заданного эллипса 100000 пар точек и, используя все эти пары точек, найдите ответы, округляя их до одного знака после запятой.

    Args:
        a (numerical): 1-я полуось
        b (numerical): 2-я полуось
        le (numerical): Чтсло из первого условия(A)
        sign (numerical): '>' или '<' чем 0

    ## Prints
        `answer` каждое значение по очереди.<br>C запятой вместо точки и сокращением до соответствующего количества знаков после запятой

    Returns:
        `answer` (tuple): Соответствующие величины
    """
    import scipy.stats
    import numpy
    # Вероятность P(A)


    U = scipy.stats.uniform(0, 2*a)
    V = scipy.stats.uniform(0, 2*b)
    N = 100_000
    count = 0
    points = []
    
    signs = {'>':max,'<':min}

    for i in range(N):
        x = U.rvs(size=1)[0]
        y = V.rvs(size=1)[0]
        if (x-a)**2/a**2 + (y-b)**2/b**2 < 1:
            points.append((x, y))

    for i in range(N):
        point_1 = numpy.random.choice(points)
        point_2 = numpy.random.choice(points)
        if ((point_1[0] - point_2[0])**2 + (point_1[1] - point_2[1])**2)**0.5 < le:
            count += 1

    answer = [count/N]
    
    print('Вероятность P(A) = ' + rrstr(answer[0],1))
    
    # Вероятность P(A|B)

    U = scipy.stats.uniform(0, 2*a)
    V = scipy.stats.uniform(0, 2*b)
    N = 100_000
    count1 = 0
    count2 = 0
    points = []

    for i in range(N):
        x = U.rvs(size=1)[0]
        y = V.rvs(size=1)[0]
        if (x-a)**2/a**2 + (y-b)**2/b**2 < 1:
            points.append((x, y))

    for i in range(N):
        point_1 = numpy.random.choice(points)
        point_2 = numpy.random.choice(points)
        
        if ((point_1[0] - point_2[0])**2 + (point_1[1] - point_2[1])**2)**0.5 < le and signs[sign](point_1[0],a)==point_1[0] and signs[sign](point_2[0], a)==point_2[0] and signs[sign](point_1[1], b)==point_1[1] and signs[sign](point_2[1],b)==point_2[1]:
            count1 += 1
        if signs[sign](point_1[0],a)==point_1[0] and signs[sign](point_2[0], a)==point_2[0] and signs[sign](point_1[1], b)==point_1[1] and signs[sign](point_2[1],b)==point_2[1]:
            count2 += 1
            
    answer.append(count1/count2)
    print('Условную вероятность P(A|B) = ' + rrstr(answer[1],1))
    
    answer = tuple(answer)
    return answer
#######################################################################################################################
def ACMK_3(a,b,le):
    """В области, ограниченной эллипсом (u/`a`)^2+(v/`b`)^2=1,
    случайным образом выбираются две точки.
    Пусть A и B - события, состоящие в том, что: 
    A - расстояние между выбранными точками меньше `le`; 
    B – координаты первой точки больше 0, а координаты второй точки меньше 0.
    Найдите приближенно, методом Монте-Карло: 
    - вероятность P(A)
    - условную вероятность P(A|B)
    
    Указание: получите внутри заданного эллипса 100000 пар точек и, используя все эти пары точек, найдите ответы, округляя их до одного знака после запятой.


    Args:
        a (numerical): 1-я полуось
        b (numerical): 2-я полуось
        le (numerical): Чтсло из первого условия(A)

    ## Prints
        `answer` каждое значение по очереди.<br>C запятой вместо точки и сокращением до соответствующего количества знаков после запятой

    Returns:
        `answer` (tuple): Соответствующие величины
    """
    ''''''
    import scipy.stats
    import numpy
    # Вероятность P(A)

    U = scipy.stats.uniform(0, 2*a)
    V = scipy.stats.uniform(0, 2*b)
    N = 100_000
    count = 0
    points = []

    for i in range(N):
        x = U.rvs(size=1)[0]
        y = V.rvs(size=1)[0]
        if (x-a)**2/a**2 + (y-b)**2/b**2 < 1:
            points.append((x, y))

    for i in range(N):
        point_1 = numpy.random.choice(points)
        point_2 = numpy.random.choice(points)
        if ((point_1[0] - point_2[0])**2 + (point_1[1] - point_2[1])**2)**0.5 < le:
            count += 1

    answer = [count/N]
    
    print('Вероятность P(A) = ' + rrstr(answer[0],1))
    # Вероятность P(A|B)

    U = scipy.stats.uniform(0, 2*a)
    V = scipy.stats.uniform(0, 2*b)
    N = 100_000
    count1 = 0
    count2 = 0
    points = []

    for i in range(N):
        x = U.rvs(size=1)[0]
        y = V.rvs(size=1)[0]
        if (x-a)**2/a**2 + (y-b)**2/b**2 < 1:
            points.append((x, y))


    for i in range(N):
        point_1 = numpy.random.choice(points)
        point_2 = numpy.random.choice(points)
        if ((point_1[0] - point_2[0])**2 + (point_1[1] - point_2[1])**2)**0.5\
        and (point_1[0] > a and point_2[0] < a and point_1[1] > b and point_2[1] < b\
        or point_1[0] < a and point_2[0] > a and point_1[1] < b and point_2[1] > b):
            count1 += 1
        if point_1[0] > a and point_2[0] < a and point_1[1] > b and point_2[1] < b\
        or point_1[0] < a and point_2[0] > a and point_1[1] < b and point_2[1] > b:
            count2 += 1

    answer.append(count1/count2)
    print('Условную вероятность P(A|B) = ' + rrstr(answer[1],1))
    
    answer = tuple(answer)
    return answer
#######################################################################################################################
def ACMK_4(cond1,r,cond2,s):
    """В кубе объема `V` случайным образом выбираются точки A, B и C.
    Пусть R, S и T - события, состоящие в том, что:
    R - `cond1` угол в треугольнике ABC меньше `r`°;
    S - `cond2` угол в треугольнике ABC меньше `s`°;
    T - треугольник ABC остроугольный.
    Найдите приближенно, методом Монте-Карло:
    - условную вероятность P(R|T)
    - условную вероятность P(S|T)
    
    Указание: получите 100000 остроугольных треугольников ABC и,
    используя все эти треугольники, найдите ответы, округляя их до одного знака после запятой.

    Args:
        cond1 (str): Параметр условия R из <br> `['наименьший','наибольший','найдется','все']`
        r (numerical): Число градусов из условия R
        cond2 (str): Параметр условия S из <br> `['наименьший','наибольший','найдется','все']`
        s (numerical): Число градусов из условия S
        

    ## Prints
        `answer` каждое значение по очереди.<br>C запятой вместо точки и сокращением до соответствующего количества знаков после запятой

    Returns:
        `answer` (tuple): Соответствующие величины
    """
    import scipy.stats
    import math
            
    
    variants = {
        'наименьший': min,
        'наибольший': max,
        'найдется' : any,
        'все' : all
    }
    
    degs=[r,s]
    v_keys = list(variants.keys())
    
    conds=[cond1,cond2]
    
    X = scipy.stats.uniform()
    Y = scipy.stats.uniform()
    Z = scipy.stats.uniform()
    N = 100_000
    
    count1 = 0
    count23 = [0,0]

    for i in range(N):
        A = X.rvs(size = 3)
        B = Y.rvs(size = 3)
        C = Z.rvs(size = 3)
        
        AB = ((B[0]-A[0])**2 + (B[1]-A[1])**2 + (B[2]-A[2])**2)**0.5
        AC = ((C[0]-A[0])**2 + (C[1]-A[1])**2 + (C[2]-A[2])**2)**0.5
        BC = ((C[0]-B[0])**2 + (C[1]-B[1])**2 + (C[2]-B[2])**2)**0.5
        
        min_side = min(AB, AC, BC)
        med_side = AB + BC + AC - max(AB, AC, BC) - min(AB, AC, BC)
        max_side = max(AB, AC, BC)
        
        min_angle = math.degrees(math.acos((max_side**2 + med_side**2 - min_side**2)/(2 * med_side * max_side)))
        med_angle = math.degrees(math.acos((max_side**2 + min_side**2 - med_side**2)/(2 * min_side * max_side)))
        max_angle = math.degrees(math.acos((min_side**2 + med_side**2 - max_side**2)/(2 * med_side * min_side)))
        
        check_traingle = min_side**2 + med_side**2 > max_side**2

        # Вероятность P(T)

        if check_traingle:
            count1 += 1

        # Вероятность P(R*T)
        for c in range(len(conds)):
            if conds[c] in v_keys[:2]:
                if check_traingle and variants[conds[c]]([min_angle,med_angle,max_angle]) < degs[c]:
                    count23[c] += 1
            else:
                if check_traingle and variants[conds[c]]([min_angle < degs[c],med_angle < degs[c],max_angle < degs[c]]) :
                    count23[c] += 1

    PRT = count23[0]/count1
    PST = count23[1]/count1
    
    answer = (PRT, PST)
    
    print('Условная вероятность P(R|T) = ' + rrstr(answer[0],1))
    print('Условная вероятность P(S|T) = ' + rrstr(answer[1],1))

    return answer
#######################################################################################################################
ACMK = [ACMK_1,ACMK_2,ACMK_3,ACMK_4]


#######################################################################################################################
# Доверительный интервал для коэффициента корреляции
#######################################################################################################################
def CICC_1(x,y,trust):
    """Пусть (x1,y1);...;(x31,y31)– реализация случайной выборки (X1,Y1);...;(X31,Y31)из двумерного нормального распределения N(μx;μy;σ2x;σ2y;ρ). 
    Используя векторы x⃗ =(x1;...;x31)и y⃗ =(y1;...;y31), постройте асимптотический 0,93-доверительный интервал (θˆ1;θˆ2)для коэффициента корреляции ρ. 
    В ответе укажите: 
    - выборочный коэффициент корреляции ρˆ;
    - верхнюю границу θˆ2построенного доверительного интервала для ρ

    Args:
        x : string
            Строка вектора X, начинающийся с круглых скобок и заканчивающийся круглыми скобками. 
            x1;x2;...;x31
        y : string
            Строка вектора Y, начинающийся с круглых скобок и заканчивающийся круглыми скобками.
            y1;y2;...;y31
        trust : float
            Доверительный уровень
            trust level

    Returns
    -------
    answer : tuple
        (ρˆ, θˆ1, θˆ2)

    """
    import numpy as np
    from scipy.stats import norm

    alpha = 1 - trust

    to_array = lambda x : np.array(x[1:-1].replace(',','.').split('; ')).astype('float')
    x = to_array(x)
    y = to_array(y)

    r = np.corrcoef(x, y)[0,1]

    z = 1/2 * np.log((1 + r) / (1 - r))

    se_z = 1/np.sqrt(len(x) - 3)

    z_alpha_div_2 = norm.ppf(1 - alpha/2)

    inter = np.tanh([z+z_alpha_div_2*se_z,z-z_alpha_div_2*se_z])

    answer = (r, min(inter),max(inter))
    
    print('Выборочный коэффициент корреляции: ', rrstr(answer[0],3))
    print('Верхняя граница доверительного интервала: ', rrstr(answer[2],3))
    print('Нижняя граница доверительного интервала: ', rrstr(answer[1],3))

    return answer
#######################################################################################################################
CICC = [CICC_1]


import importlib as _importlib
from .forall import *

submodules = [
    'tvims',
    'aisd',
    'econometrics',
    'ml'
]
def __dir__():
    return submodules

def __getattr__(name):
    if name in submodules:
        return _importlib.import_module(f'matplobblib.{name}')



import numpy as np
from  inspect import getsource

def one_rrstr(x,n=0): # округление до n знаков после запятой
    if n == 0:
        return str(x)
    fmt = '{:.'+str(n)+'f}'
    return fmt.format(x).replace('.',',')

rrstr = np.vectorize(one_rrstr)


from setuptools import setup, find_packages


def readme():
  with open('README.md', 'r') as f:
    return f.read()



setup(
    name='matplobblib',
    version='0.2.29',
    packages=find_packages(),
    description='Just a library for some subjects',
    author='Ackrome',
    author_email='ivansergeyevicht@gmail.com',
    url='https://github.com/Ackrome/matplobblib',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
       
)



#######################################################################################################################
# Работа с моделями
#######################################################################################################################
def make_model_outputs(endog, exog,untrust = 0.05, endog_name = 'Y', exog_name = 'X', func_endog = lambda y: y, func_exog = lambda x: x):
    """Выполняет задания с номера 2 в определенной работе.
    Вот эти задания:
    - **Исследование взаимосвязи данных показателей с помощью диаграммы рассеяния и коэффициента корреляции.** Построить график диаграммы рассеяния зависимой переменной с экзогенным фактором. Оценить коэффициент корреляции между объясняемой и объясняющей пе-ременными. Проанализировать тесноту и направление связи.
    - **Оценивание качества спецификации модели.** Выпишите полученное уравнение регрессии.  Дайте экономическую интерпретацию параметрам модели. Проверить статистическую значимость регрессии в целом. Проверить статистическую значимость оценок параметров. Оценить точность модели с помощью средней относительной ошибки аппроксимации. Сделайте выводы   качестве уравнения регрессии.
    - **Проверьте остатки на нормальное распределение одним из тестов (выбрать самостоятельно)**
    - **Проверка предпосылки теоремы Гаусса-Маркова о гомоскедастичности случайных возмущений.** Выполнить визуальный анализ гетероскедастичности с помощью графиков.  Провести поверку по одному из тестов - выбор теста зависит от выполнения п. 4. Обосновать применение предложенного вами теста. Сделать выводы. При необходимости предложить вариант корректировки гетероскедастичности.
    
    Args:
        endog (array-like): Эндогенная переменная (Y)
        exog (array-like): Экзогенная переменная (X)
        untrust (float, optional): Параметр недоверия(чаще alpha). Стандартно = 0.05.
        endog_name (str, optional): Имя эндогенной переменной (Y). Стандартно = 'Y'.
        exog_name (str, optional):  Имя экзогенной переменной (X). Стандартно = 'X'.
        func_endog (method_object, optional): Функция для изменения эндогенной переменной (Y). Стандартно = lambda y:y.
        func_exog (method_object, optional): Функция для изменения экзогенной переменной (X). Стандартно = lambda x:x.

    """
    import scipy.stats
    import sympy
    import numpy
    import seaborn
    import matplotlib.pyplot as plt
    from IPython.display import Math,Latex
    import statsmodels.api as sm
    from statsmodels.stats.diagnostic import het_goldfeldquandt
    
    def corr_type(r)->str:
        
        ans=''
        if r > 0:
            ans += 'ПРЯМАЯ '
        elif r < 0:
                ans += 'ОБРАТНАЯ '
                
        if abs(r) == 0:
                ans += 'НЕ НАБЛЮДАЕТСЯ '
        elif abs(r) < 0.3:
                ans += 'ОЧЕНЬ СЛАБАЯ '
        elif 0.3 <= abs(r) < 0.5:
                ans += 'СЛАБАЯ'
        elif 0.5 <= abs(r) < 0.7:
                ans += 'УМЕРЕННАЯ '
        elif 0.7 <= abs(r) < 1:
                ans += 'СИЛЬНАЯ '
        else:
                ans += 'ФУНКЦИОНАЛЬНАЯ '
        return ans

    
    def custom(endog, exog):        
        vectorized_func_endog = numpy.vectorize(func_endog)
        vectorized_func_exog = numpy.vectorize(func_exog)
        
        endog = vectorized_func_endog(endog)
        exog = vectorized_func_exog(exog)
        
        return endog,exog
        
    
    prev_exog = exog.copy()
    prev_endog = endog.copy()
    ##############################################################################################################################################################
    display(Math('№2'))
    
    plt.scatter(endog, exog)
    plt.xlabel(exog_name)
    plt.ylabel(endog_name)
    plt.show()
    
    with numpy.errstate(divide='ignore', invalid='ignore'):
        cor_matr = numpy.corrcoef(numpy.concatenate((endog,exog), axis = 1), rowvar=False)
        cor_matr[numpy.isnan(cor_matr)] = 0
    
    corr  = cor_matr[0,1] 
    display(Math(('Коэффициент корреляции между исходными объясняемой и объясняющей переменными равен r_{' + f'{endog_name},' + f'{exog_name}' + '} =' + f' {corr}. Вид связи - {corr_type(corr)}').replace(' ','~')))    
    ##############################################################################################################################################################
    display(Math('№3'))
    
    model_type = custom
    
    endog, exog = model_type(endog, exog)
    
    exog = sm.add_constant(exog)
    model = sm.OLS(endog,exog).fit()
    
    if model.pvalues[0]>untrust:
        print(f'Константа, при построении модели с ней, статистически незначима.({model.pvalues[0]}>{untrust})')
        model = sm.OLS(endog,prev_exog).fit()
    
    r2=model.rsquared
    params=model.params
    nobs=model.nobs
    df=model.df_model
    tvalues=model.tvalues
    pvalues=model.pvalues
    f_pvalue=model.f_pvalue
    fvalue=model.fvalue
    
    exog = model.model.exog                                  # X
    endog = model.model.endog.reshape((exog.shape[0],1))     # Y
    predict = model.predict(exog)
    
    display(model.summary())   
    
    t_crit = scipy.stats.t.ppf(1 - untrust/2, nobs - (df + 1))
    print(f'Коэффициент Детерминаци равен: {str(round(r2*100,3)).replace(".",",")}% .')
    print(f'\nКоэфициенты b при каждом члене равны соответственно: {params} .')
    
    coefs = []
    
    for i in range(len(params)):
        
        if pvalues[i]<untrust:
            print(f'\nКоэффициент {str(params[i]).replace(".",",")} стат. значим, т.к. значение t = {str(round(abs(tvalues[i]),4)).replace(".",",")} больше t_критического = {str(round(t_crit,4)).replace(".",",")} <=> pvalue={str(pvalues[i]).replace(".",",")} < {str(untrust).replace(".",",")}')
            coefs.append(params[i])
            
        else:
            print(f'\nКоэффициент {str(params[i]).replace(".",",")} стат. незначим, т.к. значение t = {str(round(abs(tvalues[i]),4)).replace(".",",")} меньше t_критического = {str(round(t_crit,4)).replace(".",",")} <=> pvalue={str(pvalues[i]).replace(".",",")} > {str(untrust).replace(".",",")}')
            coefs.append(0)
            
    if untrust>f_pvalue:
        print(f'\n\nРегрессия стат. значима, т.к. F-значение критерия фишера = {str(round(fvalue,4)).replace(".",",")} больше F_критического <=> fvalue={str(f_pvalue).replace(".",",")} < {str(untrust).replace(".",",")}')
    else:
        print(f'\n\nРегрессия стат. незначима, т.к. F-значение критерия фишера = {str(round(fvalue,4)).replace(".",",")} меньше F_критического <=> fvalue={str(f_pvalue).replace(".",",")} > {str(untrust).replace(".",",")}')
    
    
    try:
        if exog.shape[1]>=2:    
            x = [1]+[sympy.symbols(','.join(f'x_{i+1}' for i in range(len(params)-1)))]
            y = sum(numpy.multiply(x,coefs))
            
            display(Math('Уравнение~регрессии:~~~~y = '+ sympy.latex(y)))
            
        elif exog.shape[1]==1:
            x = [sympy.symbols(','.join(f'x_{i+1}' for i in range(len(params))))]
            y = sum(numpy.multiply(x,coefs))
            
            display(Math('Уравнение~регрессии:~~~~y = '+ sympy.latex(y)))
        
    except:
        print('Нет коэффициентов')
        pass

    
    display(Math(f'Средняя относительная ошибка апроксимации (MAPE) = {numpy.mean(abs(model.predict(exog) - endog)/endog)*100}\%'.replace(' ','~')))
    print('Это метрика, которая используется для оценки точности модели.\nОна показывает, насколько модель отклоняется от истинных значений в среднем в процентах.')
    coefs = numpy.array(coefs)
    tr_coefs_indexes = numpy.where(coefs != 0)

    
    concat_tr_array = numpy.concatenate((endog,exog[:,*tr_coefs_indexes]), axis = 1)
    
    with numpy.errstate(divide='ignore', invalid='ignore'):
        cor_matr = numpy.corrcoef(concat_tr_array, rowvar=False)
        cor_matr[numpy.isnan(cor_matr)] = 0 
    
    
    ticks = ['$y$']+[f'${i}$' for i in x]
    
    
    plt.scatter(endog,model.predict(exog))
    plt.plot([endog.min(),endog.max()],[endog.min(),endog.max()], c='r')
    plt.xlabel(endog_name)
    plt.ylabel(endog_name)
    plt.show()
    
    seaborn.heatmap(cor_matr, annot=True, xticklabels=ticks, yticklabels=ticks, cmap='coolwarm')
    plt.show()
    ##############################################################################################################################################################
    display(Math('№4'))
    resid = model.resid

    display(Math('Jarque-Bera'))
    jb_stat, jb_p_value = scipy.stats.jarque_bera(resid)
    print("Статистика:", jb_stat)
    print("p-значение:", jb_p_value)

    if jb_p_value < untrust:
        print("Данные не распределены нормально (отвергаем H0) \n")
    else:
        print("Данные распределены нормально (не отвергаем H0)\n")

    display(Math('Shapiro-Wilk'))
    stat, p_value = scipy.stats.shapiro(resid)

    print("Статистика:", stat)
    print("p-значение:", p_value)

    if p_value > untrust:
        print("Данные распределены нормально (не отвергаем H0)\n")
    else:
        print("Данные не распределены нормально (отвергаем H0) \n")



    def helwig_test(data):
        # Шаг 1: Сортируем данные и определяем размер выборки
        data_sorted = numpy.sort(data)
        n = len(data)

        # Шаг 2: Оценка среднего и стандартного отклонения
        mean, std = numpy.mean(data), numpy.std(data, ddof=1)

        # Шаг 3: Вычисляем эмпирическую функцию распределения (ЭФР)
        ecdf = numpy.arange(1, n + 1) / n

        # Шаг 4: Строим теоретическую нормальную функцию распределения (НФР)
        theoretical_cdf = scipy.stats.norm.cdf(data_sorted, mean, std)


        # Шаг 5: Вычисляем максимальное отклонение между ЭФР и НФР
        max_deviation = numpy.max(numpy.abs(ecdf - theoretical_cdf))

        # Вывод результата
        print("Максимальное отклонение (D):", max_deviation)
        return max_deviation


    display(Math('Helwig'))
    n = len(resid)
    
    critical_value = scipy.stats.kstwobign.ppf(1 - untrust) / numpy.sqrt(n)
    
    if helwig_test(resid) > critical_value:
        
        print(f"H0 отвергается на уровне значимости {untrust}.")
    else:
        
        print(f"Нет оснований отвергнуть H0 на уровне значимости {untrust}.")
    ##############################################################################################################################################################
    display(Math('№5'))
    
    display(Math('Goldfeld-Quandt'))
    
    k=0
    if model!='linear':
        print('Осуществим этот тест на преобразованных для модели данных')
        j_exog =  (exog[:,1] if exog.shape[1]>=2 else exog).reshape((-1,1))
        if het_goldfeldquandt(endog,j_exog)[1] < untrust:
            print(f'Согласно тесту Голдфельда-Куандта, данные гетероскедастичны: {het_goldfeldquandt(endog,j_exog)[1]} < {untrust}\n')
            k+=1
        else:
            print(f'Согласно тесту Голдфельда-Куандта, данные гомоскедастичны: {het_goldfeldquandt(endog,j_exog)[1]} > {untrust}')
        
        print('Это заметно и на графике:')
        
        plt.scatter(resid**2,j_exog)
        plt.xlabel(exog_name)
        plt.ylabel('$e^2$')
        plt.show()
        
    print('Осуществим этот тест на изначальных данных')
    if het_goldfeldquandt(prev_endog, prev_exog)[1] < untrust:
        print(f'Согласно тесту Голдфельда-Куандта, данные гетероскедастичны: {het_goldfeldquandt(prev_endog, prev_exog)[1]} < {untrust}\n')
        k+=1
    else:
        print(f'Согласно тесту Голдфельда-Куандта, данные гомоскедастичны: {het_goldfeldquandt(prev_endog, prev_exog)[1]} > {untrust}\n')
    print('Это заметно и на графике:')
        
    plt.scatter(resid**2, prev_exog)
    plt.xlabel(exog_name)
    plt.ylabel('$e^2$')
    plt.show()
    
    if k>=1:
        print(Math('Попробуем провести корректировку гетероскедастичности делением функции на x^{1/2}'.replace(' ','~')))
        
        prev_j_exog = j_exog.copy()
        j_exog = sm.add_constant(j_exog)
        j_exog = j_exog/prev_j_exog**0.5
        y_x05 = predict/prev_j_exog**0.5
        model = sm.OLS(y_x05,j_exog).fit()
        
        
        if any([model.pvalues[i]>untrust for i in range(len(model.pvalues))]) or model.f_pvalue>untrust:
            print(f'Данная модель оказалась статистически незначимой.Попробуем построить другую, делением функции на x ')
            j_exog = prev_j_exog.copy()
            j_exog = sm.add_constant(j_exog)
            j_exog = j_exog/prev_j_exog
            y_x1 = predict/prev_j_exog
            
            model = sm.OLS(y_x1,j_exog).fit()
            
            if any([model.pvalues[i]>untrust for i in range(len(model.pvalues))]) or model.f_pvalue>untrust:
                print('Попытки скорректировать гетероскедастичность не увенчались успехом')
    
    return model
#######################################################################################################################


from .data_structures import *
from  .search_algorythms import *
from .additional_funcs import *
from .. import *

data1 = text1.split('abracadabrabibidi')
data_structures_dict = dict([(x.split('\n')[1].replace("# ",''),x) for x in data1])
data2 = text2.split('abracadabrabibidi')
search_algorythms_dict = dict([(x.split('\n')[1].replace("# ",''),x) for x in data2])
data3 = text3.split('abracadabrabibidi')
additional_funcs_dict = dict([(x.split('\n')[1].replace("# ",''),x) for x in data3])

themes = {
        'Структуры данных': list(data_structures_dict.keys()),
        'Алгоритмы сортировки': list(search_algorythms_dict.keys()),
        'Дополнительные функции' : list(additional_funcs_dict.keys()),
        'Вывести функцию буфера обмена': ['enable_ppc'],
              }
# Тема -> Выбор структуры -> структура
def description(dict_to_show = themes, key=None, show_only_keys:bool = True):
    if dict_to_show=='Вывести функцию буфера обмена':
            return print(enable_ppc)
        
    elif type(dict_to_show) == str and dict_to_show!='Вывести функцию буфера обмена' and key==None:
            dict_to_show = {
            'Структуры данных': data_structures_dict,
            'Алгоритмы сортировки': search_algorythms_dict,
            'Дополнительные функции' : additional_funcs_dict,
                }[dict_to_show]
            
            text = ""
            length1=1+max([len(x) for x in list(dict_to_show.keys())])
            for key in dict_to_show.keys():
                text += f'{key:<{length1}}\n'
                
            return print(text)
        
    elif type(dict_to_show) == str and dict_to_show!='Вывести функцию буфера обмена' and key in list({
            'Структуры данных': data_structures_dict,
            'Алгоритмы сортировки': search_algorythms_dict,
            'Дополнительные функции' : additional_funcs_dict,
                }[dict_to_show].keys()):
        
        return print({
            'Структуры данных': data_structures_dict,
            'Алгоритмы сортировки': search_algorythms_dict,
            'Дополнительные функции' : additional_funcs_dict,
                }[dict_to_show][key])
    else:
        show_only_keys=False
    text = ""
    length1=1+max([len(x) for x in list(dict_to_show.keys())])
    for key in dict_to_show.keys():
        text += f'{key:^{length1}}'
        if not show_only_keys:
            text +=': '
            for f in dict_to_show[key]:
                text += f'{f};\n'+' '*(length1+2)
        text += '\n'
    print(text)

def enable_ppc():
    return '''import pyperclip

#Делаем функцию которая принимает переменную text
def write(name):
    pyperclip.copy(name) #Копирует в буфер обмена информацию
    pyperclip.paste()'''
